#![feature(if_let_guard, let_chains)]

pub mod cpu;
pub mod device;
pub mod rt;

pub use crate::cpu::*;
pub use crate::device::*;
pub use crate::rt::*;

use hercules_ir::*;

/*
 * Basic block info consists of two things:
 *
 * 1. A map from node to block (named by control nodes).
 * 2. For each node, which nodes are in its own block.
 *
 * Note that for #2, the structure is Vec<NodeID>, meaning the nodes are ordered
 * inside the block. This order corresponds to the traversal order of the nodes
 * in the block needed by the backend code generators.
 */
pub type BasicBlocks = (Vec<NodeID>, Vec<Vec<NodeID>>);

/*
 * The alignment of a type does not depend on dynamic constants.
 */
pub fn get_type_alignment(types: &Vec<Type>, ty: TypeID) -> usize {
    match types[ty.idx()] {
        Type::Control => panic!(),
        Type::Boolean | Type::Integer8 | Type::UnsignedInteger8 => 1,
        Type::Integer16 | Type::UnsignedInteger16 => 2,
        Type::Integer32 | Type::UnsignedInteger32 | Type::Float32 => 4,
        Type::Integer64 | Type::UnsignedInteger64 | Type::Float64 => 8,
        Type::Product(ref members) | Type::Summation(ref members) => members
            .into_iter()
            .map(|id| get_type_alignment(types, *id))
            .max()
            .unwrap_or(1),
        Type::Array(elem, _) => get_type_alignment(types, elem),
    }
}
