use std::collections::{HashMap, HashSet};

use bitvec::prelude::*;

use crate::*;

/*
 * Top level function for creating a fork-join map. Map is from fork node ID to
 * join node ID, since a join can easily determine the fork it corresponds to
 * (that's the mechanism used to implement this analysis). This analysis depends
 * on type information.
 */
pub fn fork_join_map(function: &Function, control: &Subgraph) -> HashMap<NodeID, NodeID> {
    let mut fork_join_map = HashMap::new();
    for idx in 0..function.nodes.len() {
        // We only care about join nodes.
        if function.nodes[idx].is_join() {
            // Iterate the control predecessors until finding a fork. Maintain a
            // counter of unmatched fork-join pairs seen on the way, since fork-
            // joins may be nested. Every join is dominated by their fork, so
            // just iterate the first unseen predecessor of each control node.
            let join_id = NodeID::new(idx);
            let mut unpaired = 0;
            let mut cursor = join_id;
            let mut seen = HashSet::<NodeID>::new();
            let fork_id = loop {
                cursor = control
                    .preds(cursor)
                    .filter(|pred| !seen.contains(pred))
                    .next()
                    .unwrap();
                seen.insert(cursor);

                if function.nodes[cursor.idx()].is_join() {
                    unpaired += 1;
                } else if function.nodes[cursor.idx()].is_fork() && unpaired > 0 {
                    unpaired -= 1;
                } else if function.nodes[cursor.idx()].is_fork() {
                    break cursor;
                }
            };
            fork_join_map.insert(fork_id, join_id);
        }
    }
    fork_join_map
}

/*
 * Find fork/join nests that each control node is inside of. Result is a map
 * from each control node to a list of fork nodes. The fork nodes are listed in
 * ascending order of nesting.
 */
pub fn compute_fork_join_nesting(
    function: &Function,
    dom: &DomTree,
    fork_join_map: &HashMap<NodeID, NodeID>,
) -> HashMap<NodeID, Vec<NodeID>> {
    // For each control node, ascend dominator tree, looking for fork nodes. For
    // each fork node, make sure each control node isn't strictly dominated by
    // the corresponding join node.
    (0..function.nodes.len())
        .map(NodeID::new)
        .filter(|id| dom.contains(*id))
        .map(|id| {
            (
                id,
                dom.ascend(id)
                    // Filter for forks that dominate this control node,
                    .filter(|id| function.nodes[id.idx()].is_fork())
                    // where its corresponding join doesn't dominate the control
                    // node (if so, then this control is after the fork-join).
                    .filter(|fork_id| !dom.does_prop_dom(fork_join_map[&fork_id], id))
                    .collect(),
            )
        })
        .collect()
}

/*
 * Check if a data node dominates a control node. This involves checking all
 * immediate control uses to see if they dominate the queried control node.
 */
pub fn does_data_dom_control(
    function: &Function,
    data: NodeID,
    control: NodeID,
    dom: &DomTree,
) -> bool {
    let mut stack = vec![data];
    let mut visited = bitvec![u8, Lsb0; 0; function.nodes.len()];
    visited.set(data.idx(), true);

    while let Some(pop) = stack.pop() {
        let node = &function.nodes[pop.idx()];

        let imm_control = match node {
            Node::Phi { control, data: _ }
            | Node::Reduce {
                control,
                init: _,
                reduct: _,
            }
            | Node::Call {
                control,
                function: _,
                dynamic_constants: _,
                args: _,
            } => Some(*control),
            _ if node.is_control() => Some(pop),
            _ => {
                for u in get_uses(node).as_ref() {
                    if !visited[u.idx()] {
                        visited.set(u.idx(), true);
                        stack.push(*u);
                    }
                }
                None
            }
        };

        if let Some(imm_control) = imm_control
            && !dom.does_dom(imm_control, control)
        {
            return false;
        }
    }

    true
}
