use std::cell::RefCell;
use std::collections::{HashMap, HashSet};
use std::fs::File;
use std::io::Write;
use std::iter::zip;
use std::process::{Command, Stdio};

use serde::Deserialize;

use tempfile::TempDir;

use hercules_cg::*;
use hercules_ir::*;

use crate::*;

/*
 * Passes that can be run on a module.
 */
#[derive(Debug, Clone, Deserialize)]
pub enum Pass {
    DCE,
    CCP,
    GVN,
    PhiElim,
    Forkify,
    ForkGuardElim,
    SLF,
    WritePredication,
    Predication,
    SROA,
    Inline,
    Outline,
    InterproceduralSROA,
    DeleteUncalled,
    ForkSplit,
    Unforkify,
    InferSchedules,
    GCM,
    FloatCollections,
    Verify,
    // Parameterized over whether analyses that aid visualization are necessary.
    // Useful to set to false if displaying a potentially broken module.
    Xdot(bool),
    // Parameterized over output directory and module name.
    Codegen(String, String),
    // Parameterized over where to serialize module to.
    Serialize(String),
}

/*
 * Manages passes to be run on an IR module. Transparently handles analysis
 * requirements for optimizations.
 */
#[derive(Debug, Clone)]
pub struct PassManager {
    module: Module,

    // Passes to run.
    passes: Vec<Pass>,

    // Cached analysis results.
    pub def_uses: Option<Vec<ImmutableDefUseMap>>,
    pub reverse_postorders: Option<Vec<Vec<NodeID>>>,
    pub typing: Option<ModuleTyping>,
    pub control_subgraphs: Option<Vec<Subgraph>>,
    pub doms: Option<Vec<DomTree>>,
    pub postdoms: Option<Vec<DomTree>>,
    pub fork_join_maps: Option<Vec<HashMap<NodeID, NodeID>>>,
    pub fork_join_nests: Option<Vec<HashMap<NodeID, Vec<NodeID>>>>,
    pub loops: Option<Vec<LoopTree>>,
    pub reduce_cycles: Option<Vec<HashMap<NodeID, HashSet<NodeID>>>>,
    pub data_nodes_in_fork_joins: Option<Vec<HashMap<NodeID, HashSet<NodeID>>>>,
    pub bbs: Option<Vec<BasicBlocks>>,
    pub collection_objects: Option<CollectionObjects>,
    pub callgraph: Option<CallGraph>,
}

impl PassManager {
    pub fn new(module: Module) -> Self {
        PassManager {
            module,
            passes: vec![],
            def_uses: None,
            reverse_postorders: None,
            typing: None,
            control_subgraphs: None,
            doms: None,
            postdoms: None,
            fork_join_maps: None,
            fork_join_nests: None,
            loops: None,
            reduce_cycles: None,
            data_nodes_in_fork_joins: None,
            bbs: None,
            collection_objects: None,
            callgraph: None,
        }
    }

    pub fn add_pass(&mut self, pass: Pass) {
        self.passes.push(pass);
    }

    pub fn make_def_uses(&mut self) {
        if self.def_uses.is_none() {
            self.def_uses = Some(self.module.functions.iter().map(def_use).collect());
        }
    }

    pub fn make_reverse_postorders(&mut self) {
        if self.reverse_postorders.is_none() {
            self.make_def_uses();
            self.reverse_postorders = Some(
                self.def_uses
                    .as_ref()
                    .unwrap()
                    .iter()
                    .map(reverse_postorder)
                    .collect(),
            );
        }
    }

    pub fn make_typing(&mut self) {
        if self.typing.is_none() {
            self.make_reverse_postorders();
            self.typing = Some(
                typecheck(&mut self.module, self.reverse_postorders.as_ref().unwrap()).unwrap(),
            );
        }
    }

    pub fn make_control_subgraphs(&mut self) {
        if self.control_subgraphs.is_none() {
            self.make_def_uses();
            self.control_subgraphs = Some(
                zip(&self.module.functions, self.def_uses.as_ref().unwrap())
                    .map(|(function, def_use)| control_subgraph(function, def_use))
                    .collect(),
            );
        }
    }

    pub fn make_doms(&mut self) {
        if self.doms.is_none() {
            self.make_control_subgraphs();
            self.doms = Some(
                self.control_subgraphs
                    .as_ref()
                    .unwrap()
                    .iter()
                    .map(|subgraph| dominator(subgraph, NodeID::new(0)))
                    .collect(),
            );
        }
    }

    pub fn make_postdoms(&mut self) {
        if self.postdoms.is_none() {
            self.make_control_subgraphs();
            self.postdoms = Some(
                zip(
                    self.control_subgraphs.as_ref().unwrap().iter(),
                    self.module.functions.iter(),
                )
                .map(|(subgraph, function)| dominator(subgraph, NodeID::new(function.nodes.len())))
                .collect(),
            );
        }
    }

    pub fn make_fork_join_maps(&mut self) {
        if self.fork_join_maps.is_none() {
            self.make_control_subgraphs();
            self.fork_join_maps = Some(
                zip(
                    self.module.functions.iter(),
                    self.control_subgraphs.as_ref().unwrap().iter(),
                )
                .map(|(function, subgraph)| fork_join_map(function, subgraph))
                .collect(),
            );
        }
    }

    pub fn make_fork_join_nests(&mut self) {
        if self.fork_join_nests.is_none() {
            self.make_doms();
            self.make_fork_join_maps();
            self.fork_join_nests = Some(
                zip(
                    self.module.functions.iter(),
                    zip(
                        self.doms.as_ref().unwrap().iter(),
                        self.fork_join_maps.as_ref().unwrap().iter(),
                    ),
                )
                .map(|(function, (dom, fork_join_map))| {
                    compute_fork_join_nesting(function, dom, fork_join_map)
                })
                .collect(),
            );
        }
    }

    pub fn make_loops(&mut self) {
        if self.loops.is_none() {
            self.make_control_subgraphs();
            self.make_doms();
            self.make_fork_join_maps();
            let control_subgraphs = self.control_subgraphs.as_ref().unwrap().iter();
            let doms = self.doms.as_ref().unwrap().iter();
            let fork_join_maps = self.fork_join_maps.as_ref().unwrap().iter();
            self.loops = Some(
                zip(control_subgraphs, zip(doms, fork_join_maps))
                    .map(|(control_subgraph, (dom, fork_join_map))| {
                        loops(control_subgraph, NodeID::new(0), dom, fork_join_map)
                    })
                    .collect(),
            );
        }
    }

    pub fn make_reduce_cycles(&mut self) {
        if self.reduce_cycles.is_none() {
            self.make_def_uses();
            let def_uses = self.def_uses.as_ref().unwrap().iter();
            self.reduce_cycles = Some(
                zip(self.module.functions.iter(), def_uses)
                    .map(|(function, def_use)| reduce_cycles(function, def_use))
                    .collect(),
            );
        }
    }

    pub fn make_data_nodes_in_fork_joins(&mut self) {
        if self.data_nodes_in_fork_joins.is_none() {
            self.make_def_uses();
            self.make_fork_join_maps();
            self.data_nodes_in_fork_joins = Some(
                zip(
                    self.module.functions.iter(),
                    zip(
                        self.def_uses.as_ref().unwrap().iter(),
                        self.fork_join_maps.as_ref().unwrap().iter(),
                    ),
                )
                .map(|(function, (def_use, fork_join_map))| {
                    data_nodes_in_fork_joins(function, def_use, fork_join_map)
                })
                .collect(),
            );
        }
    }

    pub fn make_collection_objects(&mut self) {
        if self.collection_objects.is_none() {
            self.make_reverse_postorders();
            self.make_typing();
            self.make_callgraph();
            let reverse_postorders = self.reverse_postorders.as_ref().unwrap();
            let typing = self.typing.as_ref().unwrap();
            let callgraph = self.callgraph.as_ref().unwrap();
            self.collection_objects = Some(collection_objects(
                &self.module,
                reverse_postorders,
                typing,
                callgraph,
            ));
        }
    }

    pub fn make_callgraph(&mut self) {
        if self.callgraph.is_none() {
            self.callgraph = Some(callgraph(&self.module));
        }
    }

    pub fn run_passes(&mut self) {
        for pass in self.passes.clone().iter() {
            match pass {
                Pass::DCE => {
                    self.make_def_uses();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        dce(&mut editor);

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::InterproceduralSROA => {
                    self.make_def_uses();
                    self.make_typing();

                    let constants_ref = RefCell::new(std::mem::take(&mut self.module.constants));
                    let dynamic_constants_ref =
                        RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                    let types_ref = RefCell::new(std::mem::take(&mut self.module.types));

                    let def_uses = self.def_uses.as_ref().unwrap();

                    let mut editors: Vec<_> = self
                        .module
                        .functions
                        .iter_mut()
                        .enumerate()
                        .map(|(i, f)| {
                            FunctionEditor::new(
                                f,
                                FunctionID::new(i),
                                &constants_ref,
                                &dynamic_constants_ref,
                                &types_ref,
                                &def_uses[i],
                            )
                        })
                        .collect();

                    interprocedural_sroa(&mut editors);

                    self.module.constants = constants_ref.take();
                    self.module.dynamic_constants = dynamic_constants_ref.take();
                    self.module.types = types_ref.take();

                    for func in self.module.functions.iter_mut() {
                        func.delete_gravestones();
                    }

                    self.clear_analyses();
                }
                Pass::CCP => {
                    self.make_def_uses();
                    self.make_reverse_postorders();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let reverse_postorders = self.reverse_postorders.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        ccp(&mut editor, &reverse_postorders[idx]);

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::GVN => {
                    self.make_def_uses();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        gvn(&mut editor, false);

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::Forkify => {
                    self.make_def_uses();
                    self.make_loops();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let loops = self.loops.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        forkify(
                            &mut self.module.functions[idx],
                            &self.module.constants,
                            &mut self.module.dynamic_constants,
                            &def_uses[idx],
                            &loops[idx],
                        );
                        let num_nodes = self.module.functions[idx].nodes.len();
                        self.module.functions[idx]
                            .schedules
                            .resize(num_nodes, vec![]);
                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::PhiElim => {
                    self.make_def_uses();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        phi_elim(&mut editor);

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::ForkGuardElim => {
                    self.make_def_uses();
                    self.make_fork_join_maps();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let fork_join_maps = self.fork_join_maps.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        fork_guard_elim(
                            &mut self.module.functions[idx],
                            &self.module.constants,
                            &fork_join_maps[idx],
                            &def_uses[idx],
                        );
                        let num_nodes = self.module.functions[idx].nodes.len();
                        self.module.functions[idx]
                            .schedules
                            .resize(num_nodes, vec![]);
                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::SLF => {
                    self.make_def_uses();
                    self.make_reverse_postorders();
                    self.make_typing();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let reverse_postorders = self.reverse_postorders.as_ref().unwrap();
                    let typing = self.typing.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        slf(&mut editor, &reverse_postorders[idx], &typing[idx]);

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        println!("{}", self.module.functions[idx].name);
                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::WritePredication => {
                    self.make_def_uses();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        write_predication(&mut editor);

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::Predication => {
                    self.make_def_uses();
                    self.make_typing();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let typing = self.typing.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        predication(&mut editor, &typing[idx]);

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::SROA => {
                    self.make_def_uses();
                    self.make_reverse_postorders();
                    self.make_typing();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let reverse_postorders = self.reverse_postorders.as_ref().unwrap();
                    let typing = self.typing.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        sroa(&mut editor, &reverse_postorders[idx], &typing[idx]);

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::Inline => {
                    self.make_def_uses();
                    self.make_callgraph();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let callgraph = self.callgraph.as_ref().unwrap();
                    let constants_ref = RefCell::new(std::mem::take(&mut self.module.constants));
                    let dynamic_constants_ref =
                        RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                    let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                    let mut editors: Vec<_> = zip(
                        self.module.functions.iter_mut().enumerate(),
                        def_uses.iter(),
                    )
                    .map(|((idx, func), def_use)| {
                        FunctionEditor::new(
                            func,
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            def_use,
                        )
                    })
                    .collect();
                    inline(&mut editors, callgraph);

                    self.module.constants = constants_ref.take();
                    self.module.dynamic_constants = dynamic_constants_ref.take();
                    self.module.types = types_ref.take();

                    for func in self.module.functions.iter_mut() {
                        func.delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::Outline => {
                    self.make_def_uses();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let constants_ref = RefCell::new(std::mem::take(&mut self.module.constants));
                    let dynamic_constants_ref =
                        RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                    let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                    let old_num_funcs = self.module.functions.len();
                    let mut editors: Vec<_> = zip(
                        self.module.functions.iter_mut().enumerate(),
                        def_uses.iter(),
                    )
                    .map(|((idx, func), def_use)| {
                        FunctionEditor::new(
                            func,
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            def_use,
                        )
                    })
                    .collect();
                    for editor in editors.iter_mut() {
                        collapse_returns(editor);
                        ensure_between_control_flow(editor);
                    }
                    self.module.constants = constants_ref.take();
                    self.module.dynamic_constants = dynamic_constants_ref.take();
                    self.module.types = types_ref.take();
                    self.clear_analyses();

                    self.make_def_uses();
                    self.make_typing();
                    self.make_control_subgraphs();
                    self.make_doms();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let typing = self.typing.as_ref().unwrap();
                    let control_subgraphs = self.control_subgraphs.as_ref().unwrap();
                    let doms = self.doms.as_ref().unwrap();
                    let constants_ref = RefCell::new(std::mem::take(&mut self.module.constants));
                    let dynamic_constants_ref =
                        RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                    let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                    let mut editors: Vec<_> = zip(
                        self.module.functions.iter_mut().enumerate(),
                        def_uses.iter(),
                    )
                    .map(|((idx, func), def_use)| {
                        FunctionEditor::new(
                            func,
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            def_use,
                        )
                    })
                    .collect();
                    let mut new_funcs = vec![];
                    for (idx, editor) in editors.iter_mut().enumerate() {
                        let new_func_id = FunctionID::new(old_num_funcs + new_funcs.len());
                        let new_func = dumb_outline(
                            editor,
                            &typing[idx],
                            &control_subgraphs[idx],
                            &doms[idx],
                            new_func_id,
                        );
                        if let Some(new_func) = new_func {
                            new_funcs.push(new_func);
                        }
                    }
                    self.module.constants = constants_ref.take();
                    self.module.dynamic_constants = dynamic_constants_ref.take();
                    self.module.types = types_ref.take();

                    for func in self.module.functions.iter_mut() {
                        func.delete_gravestones();
                    }
                    self.module.functions.extend(new_funcs);
                    self.clear_analyses();
                }
                Pass::DeleteUncalled => {
                    self.make_def_uses();
                    self.make_callgraph();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let callgraph = self.callgraph.as_ref().unwrap();
                    let constants_ref = RefCell::new(std::mem::take(&mut self.module.constants));
                    let dynamic_constants_ref =
                        RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                    let types_ref = RefCell::new(std::mem::take(&mut self.module.types));

                    // By default in an editor all nodes are mutable, which is desired in this case
                    // since we are only modifying the IDs of functions that we call.
                    let mut editors: Vec<_> = zip(
                        self.module.functions.iter_mut().enumerate(),
                        def_uses.iter(),
                    )
                    .map(|((idx, func), def_use)| {
                        FunctionEditor::new(
                            func,
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            def_use,
                        )
                    })
                    .collect();

                    let new_idx = delete_uncalled(&mut editors, callgraph);
                    self.module.constants = constants_ref.take();
                    self.module.dynamic_constants = dynamic_constants_ref.take();
                    self.module.types = types_ref.take();

                    for func in self.module.functions.iter_mut() {
                        func.delete_gravestones();
                    }

                    self.fix_deleted_functions(&new_idx);
                    self.clear_analyses();

                    assert!(self.module.functions.len() > 0, "PANIC: There are no entry functions in the Hercules module being compiled, and they all got deleted by DeleteUncalled. Please mark at least one function as an entry!");
                }
                Pass::ForkSplit => {
                    self.make_def_uses();
                    self.make_fork_join_maps();
                    self.make_reduce_cycles();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let fork_join_maps = self.fork_join_maps.as_ref().unwrap();
                    let reduce_cycles = self.reduce_cycles.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        fork_split(&mut editor, &fork_join_maps[idx], &reduce_cycles[idx]);

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::Unforkify => {
                    self.make_def_uses();
                    self.make_fork_join_maps();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let fork_join_maps = self.fork_join_maps.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        unforkify(&mut editor, &fork_join_maps[idx]);

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::GCM => loop {
                    self.make_def_uses();
                    self.make_reverse_postorders();
                    self.make_typing();
                    self.make_control_subgraphs();
                    self.make_doms();
                    self.make_fork_join_maps();
                    self.make_loops();
                    self.make_collection_objects();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let reverse_postorders = self.reverse_postorders.as_ref().unwrap();
                    let typing = self.typing.as_ref().unwrap();
                    let doms = self.doms.as_ref().unwrap();
                    let fork_join_maps = self.fork_join_maps.as_ref().unwrap();
                    let loops = self.loops.as_ref().unwrap();
                    let control_subgraphs = self.control_subgraphs.as_ref().unwrap();
                    let collection_objects = self.collection_objects.as_ref().unwrap();
                    let mut bbs = vec![];
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        if let Some(bb) = gcm(
                            &mut editor,
                            &def_uses[idx],
                            &reverse_postorders[idx],
                            &typing[idx],
                            &control_subgraphs[idx],
                            &doms[idx],
                            &fork_join_maps[idx],
                            &loops[idx],
                            collection_objects,
                        ) {
                            bbs.push(bb);
                        }

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                    if bbs.len() == self.module.functions.len() {
                        self.bbs = Some(bbs);
                        break;
                    }
                },
                Pass::FloatCollections => {
                    self.make_def_uses();
                    self.make_typing();
                    self.make_callgraph();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let typing = self.typing.as_ref().unwrap();
                    let callgraph = self.callgraph.as_ref().unwrap();
                    let devices = device_placement(&self.module.functions, &callgraph);
                    let constants_ref = RefCell::new(std::mem::take(&mut self.module.constants));
                    let dynamic_constants_ref =
                        RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                    let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                    let mut editors: Vec<_> = zip(
                        self.module.functions.iter_mut().enumerate(),
                        def_uses.iter(),
                    )
                    .map(|((idx, func), def_use)| {
                        FunctionEditor::new(
                            func,
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            def_use,
                        )
                    })
                    .collect();
                    float_collections(&mut editors, typing, callgraph, &devices);

                    self.module.constants = constants_ref.take();
                    self.module.dynamic_constants = dynamic_constants_ref.take();
                    self.module.types = types_ref.take();

                    for func in self.module.functions.iter_mut() {
                        func.delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::InferSchedules => {
                    self.make_def_uses();
                    self.make_fork_join_maps();
                    self.make_reduce_cycles();
                    let def_uses = self.def_uses.as_ref().unwrap();
                    let fork_join_maps = self.fork_join_maps.as_ref().unwrap();
                    let reduce_cycles = self.reduce_cycles.as_ref().unwrap();
                    for idx in 0..self.module.functions.len() {
                        let constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.constants));
                        let dynamic_constants_ref =
                            RefCell::new(std::mem::take(&mut self.module.dynamic_constants));
                        let types_ref = RefCell::new(std::mem::take(&mut self.module.types));
                        let mut editor = FunctionEditor::new(
                            &mut self.module.functions[idx],
                            FunctionID::new(idx),
                            &constants_ref,
                            &dynamic_constants_ref,
                            &types_ref,
                            &def_uses[idx],
                        );
                        infer_parallel_reduce(
                            &mut editor,
                            &fork_join_maps[idx],
                            &reduce_cycles[idx],
                        );
                        infer_parallel_fork(&mut editor, &fork_join_maps[idx]);
                        infer_vectorizable(&mut editor, &fork_join_maps[idx]);
                        infer_tight_associative(&mut editor, &reduce_cycles[idx]);

                        self.module.constants = constants_ref.take();
                        self.module.dynamic_constants = dynamic_constants_ref.take();
                        self.module.types = types_ref.take();

                        self.module.functions[idx].delete_gravestones();
                    }
                    self.clear_analyses();
                }
                Pass::Verify => {
                    let (
                        def_uses,
                        reverse_postorders,
                        typing,
                        subgraphs,
                        doms,
                        postdoms,
                        fork_join_maps,
                    ) = verify(&mut self.module)
                        .expect("PANIC: Failed to verify Hercules IR module.");

                    // Verification produces a bunch of analysis results that
                    // may be useful for later passes.
                    self.def_uses = Some(def_uses);
                    self.reverse_postorders = Some(reverse_postorders);
                    self.typing = Some(typing);
                    self.control_subgraphs = Some(subgraphs);
                    self.doms = Some(doms);
                    self.postdoms = Some(postdoms);
                    self.fork_join_maps = Some(fork_join_maps);
                }
                Pass::Xdot(force_analyses) => {
                    self.make_reverse_postorders();
                    if *force_analyses {
                        self.make_doms();
                        self.make_fork_join_maps();
                    }
                    xdot_module(
                        &self.module,
                        self.reverse_postorders.as_ref().unwrap(),
                        self.doms.as_ref(),
                        self.fork_join_maps.as_ref(),
                    );
                }
                Pass::Codegen(output_dir, module_name) => {
                    self.make_typing();
                    self.make_control_subgraphs();
                    self.make_collection_objects();
                    self.make_callgraph();
                    let typing = self.typing.as_ref().unwrap();
                    let control_subgraphs = self.control_subgraphs.as_ref().unwrap();
                    let bbs = self.bbs.as_ref().unwrap();
                    let collection_objects = self.collection_objects.as_ref().unwrap();
                    let callgraph = self.callgraph.as_ref().unwrap();

                    let devices = device_placement(&self.module.functions, &callgraph);

                    let mut rust_rt = String::new();
                    let mut llvm_ir = String::new();
                    for idx in 0..self.module.functions.len() {
                        match devices[idx] {
                            Device::LLVM => cpu_codegen(
                                &self.module.functions[idx],
                                &self.module.types,
                                &self.module.constants,
                                &self.module.dynamic_constants,
                                &typing[idx],
                                &control_subgraphs[idx],
                                &bbs[idx],
                                &mut llvm_ir,
                            )
                            .unwrap(),
                            Device::AsyncRust => rt_codegen(
                                FunctionID::new(idx),
                                &self.module,
                                &typing[idx],
                                &control_subgraphs[idx],
                                &bbs[idx],
                                &collection_objects,
                                &callgraph,
                                &devices,
                                &mut rust_rt,
                            )
                            .unwrap(),
                            _ => todo!(),
                        }
                    }
                    println!("{}", llvm_ir);
                    println!("{}", rust_rt);

                    // Write the LLVM IR into a temporary file.
                    let tmp_dir = TempDir::new().unwrap();
                    let mut tmp_path = tmp_dir.path().to_path_buf();
                    tmp_path.push(format!("{}.ll", module_name));
                    println!("{}", tmp_path.display());
                    let mut file = File::create(&tmp_path)
                        .expect("PANIC: Unable to open output LLVM IR file.");
                    file.write_all(llvm_ir.as_bytes())
                        .expect("PANIC: Unable to write output LLVM IR file contents.");

                    // Compile LLVM IR into an ELF object file.
                    let output_archive = format!("{}/lib{}.a", output_dir, module_name);
                    println!("{}", output_archive);
                    let mut clang_process = Command::new("clang")
                        .arg(&tmp_path)
                        .arg("--emit-static-lib")
                        .arg("-O3")
                        .arg("-march=native")
                        .arg("-o")
                        .arg(&output_archive)
                        .stdin(Stdio::piped())
                        .stdout(Stdio::piped())
                        .spawn()
                        .expect("Error running clang. Is it installed?");
                    assert!(clang_process.wait().unwrap().success());

                    // Write the Rust runtime into a file.
                    let output_rt = format!("{}/rt_{}.hrt", output_dir, module_name);
                    println!("{}", output_rt);
                    let mut file = File::create(&output_rt)
                        .expect("PANIC: Unable to open output Rust runtime file.");
                    file.write_all(rust_rt.as_bytes())
                        .expect("PANIC: Unable to write output Rust runtime file contents.");
                }
                Pass::Serialize(output_file) => {
                    let module_contents: Vec<u8> = postcard::to_allocvec(&self.module).unwrap();
                    let mut file = File::create(&output_file)
                        .expect("PANIC: Unable to open output module file.");
                    file.write_all(&module_contents)
                        .expect("PANIC: Unable to write output module file contents.");
                }
            }
            eprintln!("Ran pass: {:?}", pass);
        }
    }

    fn clear_analyses(&mut self) {
        self.def_uses = None;
        self.reverse_postorders = None;
        self.typing = None;
        self.control_subgraphs = None;
        self.doms = None;
        self.postdoms = None;
        self.fork_join_maps = None;
        self.fork_join_nests = None;
        self.loops = None;
        self.reduce_cycles = None;
        self.data_nodes_in_fork_joins = None;
        self.bbs = None;
        self.collection_objects = None;
        self.callgraph = None;
    }

    pub fn get_module(self) -> Module {
        self.module
    }

    fn fix_deleted_functions(&mut self, id_mapping: &[Option<usize>]) {
        let mut idx = 0;

        // Rust does not like enumerate here, so use
        // idx outside as a hack to make it happy.
        self.module.functions.retain(|_| {
            idx += 1;
            id_mapping[idx - 1].is_some()
        });
    }
}
