use std::alloc::{alloc, alloc_zeroed, dealloc, Layout};
use std::marker::PhantomData;
use std::ptr::{copy_nonoverlapping, NonNull};
use std::slice::from_raw_parts;
use std::sync::atomic::{AtomicUsize, Ordering};

#[cfg(feature = "cuda")]
extern "C" {
    fn cuda_alloc(size: usize) -> *mut u8;
    fn cuda_alloc_zeroed(size: usize) -> *mut u8;
    fn cuda_dealloc(ptr: *mut u8);
    fn copy_cpu_to_cuda(dst: *mut u8, src: *mut u8, size: usize);
    fn copy_cuda_to_cpu(dst: *mut u8, src: *mut u8, size: usize);
    fn copy_cuda_to_cuda(dst: *mut u8, src: *mut u8, size: usize);
}

/*
 * Each object needs to get assigned a unique ID.
 */
static NUM_OBJECTS: AtomicUsize = AtomicUsize::new(1);

/*
 * An in-memory collection object that can be used by functions compiled by the
 * Hercules compiler. Memory objects can be in these states:
 *
 * 1. Shared CPU - the object has a shared reference to some CPU memory, usually
 *    from the programmer using the Hercules RT API.
 * 2. Exclusive CPU - the object has an exclusive reference to some CPU memory,
 *    usually from the programmer using the Hercules RT API.
 * 3. Owned CPU - the object owns some allocated CPU memory.
 * 4. Owned GPU - the object owns some allocated GPU memory.
 *
 * A single object can be in some combination of these objects at the same time.
 * Only some combinations are valid, because only some combinations are
 * reachable. Under this assumption, we can model an object's placement as a
 * state machine, where states are combinations of the aforementioned states,
 * and actions are requests on the CPU or GPU, immutably or mutably. Here's the
 * state transition table:
 *
 * Shared CPU = CS
 * Exclusive CPU = CE
 * Owned CPU = CO
 * Owned GPU = GO
 *
 *          CPU     Mut CPU     GPU     Mut GPU
 *       *---------------------------------------
 *  CS   |   CS        CO      CS,GO       GO
 *  CE   |   CE        CE      CE,GO       GO
 *  CO   |   CO        CO      CO,GO       GO
 *  GO   |   CO        CO        GO        GO
 * CS,GO | CS,GO       CO      CS,GO       GO
 * CE,GO | CE,GO       CE      CE,GO       GO
 * CO,GO | CO,GO       CO      CO,GO       GO
 *       |
 *
 * A HerculesBox cannot be cloned, because it may have be a mutable reference to
 * some CPU memory.
 */
#[derive(Debug)]
pub struct HerculesBox<'a> {
    cpu_shared: Option<NonOwned<'a>>,
    cpu_exclusive: Option<NonOwned<'a>>,
    cpu_owned: Option<Owned>,

    #[cfg(feature = "cuda")]
    cuda_owned: Option<Owned>,

    size: usize,
    id: usize,
}

#[derive(Clone, Debug)]
struct NonOwned<'a> {
    ptr: NonNull<u8>,
    offset: usize,
    _phantom: PhantomData<&'a u8>,
}

#[derive(Clone, Debug)]
struct Owned {
    ptr: NonNull<u8>,
    alloc_size: usize,
    offset: usize,
}

impl<'b, 'a: 'b> HerculesBox<'a> {
    pub fn from_slice<T>(slice: &'a [T]) -> Self {
        let ptr = unsafe { NonNull::new_unchecked(slice.as_ptr() as *mut u8) };
        let size = slice.len() * size_of::<T>();
        let id = NUM_OBJECTS.fetch_add(1, Ordering::Relaxed);
        HerculesBox {
            cpu_shared: Some(NonOwned {
                ptr,
                offset: 0,
                _phantom: PhantomData,
            }),
            cpu_exclusive: None,
            cpu_owned: None,

            #[cfg(feature = "cuda")]
            cuda_owned: None,

            size,
            id,
        }
    }

    pub fn from_slice_mut<T>(slice: &'a mut [T]) -> Self {
        let ptr = unsafe { NonNull::new_unchecked(slice.as_ptr() as *mut u8) };
        let size = slice.len() * size_of::<T>();
        let id = NUM_OBJECTS.fetch_add(1, Ordering::Relaxed);
        HerculesBox {
            cpu_shared: None,
            cpu_exclusive: Some(NonOwned {
                ptr,
                offset: 0,
                _phantom: PhantomData,
            }),
            cpu_owned: None,

            #[cfg(feature = "cuda")]
            cuda_owned: None,

            size,
            id,
        }
    }

    pub fn as_slice<T>(&'b mut self) -> &'b [T] {
        assert_eq!(self.size % size_of::<T>(), 0);
        unsafe { from_raw_parts(self.__cpu_ptr() as *const T, self.size / size_of::<T>()) }
    }

    unsafe fn get_cpu_ptr(&self) -> Option<NonNull<u8>> {
        self.cpu_owned
            .as_ref()
            .map(|obj| obj.ptr.byte_add(obj.offset))
            .or(self
                .cpu_exclusive
                .as_ref()
                .map(|obj| obj.ptr.byte_add(obj.offset)))
            .or(self
                .cpu_shared
                .as_ref()
                .map(|obj| obj.ptr.byte_add(obj.offset)))
    }

    #[cfg(feature = "cuda")]
    unsafe fn get_cuda_ptr(&self) -> Option<NonNull<u8>> {
        self.cuda_owned
            .as_ref()
            .map(|obj| obj.ptr.byte_add(obj.offset))
    }

    unsafe fn allocate_cpu(&mut self) -> NonNull<u8> {
        if let Some(obj) = self.cpu_owned.as_ref() {
            obj.ptr
        } else {
            let ptr =
                NonNull::new(alloc(Layout::from_size_align_unchecked(self.size, 16))).unwrap();
            self.cpu_owned = Some(Owned {
                ptr,
                alloc_size: self.size,
                offset: 0,
            });
            ptr
        }
    }

    #[cfg(feature = "cuda")]
    unsafe fn allocate_cuda(&mut self) -> NonNull<u8> {
        if let Some(obj) = self.cuda_owned.as_ref() {
            obj.ptr
        } else {
            let ptr = NonNull::new(cuda_alloc(self.size)).unwrap();
            self.cuda_owned = Some(Owned {
                ptr,
                alloc_size: self.size,
                offset: 0,
            });
            ptr
        }
    }

    unsafe fn deallocate_cpu(&mut self) {
        if let Some(obj) = self.cpu_owned.take() {
            dealloc(
                obj.ptr.as_ptr(),
                Layout::from_size_align_unchecked(obj.alloc_size, 16),
            );
        }
    }

    #[cfg(feature = "cuda")]
    unsafe fn deallocate_cuda(&mut self) {
        if let Some(obj) = self.cuda_owned.take() {
            cuda_dealloc(obj.ptr.as_ptr());
        }
    }

    pub unsafe fn __zeros(size: u64) -> Self {
        assert_ne!(size, 0);
        let size = size as usize;
        let id = NUM_OBJECTS.fetch_add(1, Ordering::Relaxed);
        HerculesBox {
            cpu_shared: None,
            cpu_exclusive: None,
            cpu_owned: Some(Owned {
                ptr: NonNull::new(alloc_zeroed(Layout::from_size_align_unchecked(size, 16)))
                    .unwrap(),
                alloc_size: size,
                offset: 0,
            }),

            #[cfg(feature = "cuda")]
            cuda_owned: None,

            size,
            id,
        }
    }

    pub unsafe fn __null() -> Self {
        HerculesBox {
            cpu_shared: None,
            cpu_exclusive: None,
            cpu_owned: None,

            #[cfg(feature = "cuda")]
            cuda_owned: None,

            size: 0,
            id: 0,
        }
    }

    pub unsafe fn __cpu_ptr(&mut self) -> *mut u8 {
        if let Some(ptr) = self.get_cpu_ptr() {
            return ptr.as_ptr();
        }
        #[cfg(feature = "cuda")]
        {
            let cuda_ptr = self.get_cuda_ptr().unwrap();
            let cpu_ptr = self.allocate_cpu();
            copy_cuda_to_cpu(cpu_ptr.as_ptr(), cuda_ptr.as_ptr(), self.size);
            return cpu_ptr.as_ptr();
        }
        panic!()
    }

    pub unsafe fn __cpu_ptr_mut(&mut self) -> *mut u8 {
        let cpu_ptr = self.__cpu_ptr();
        if Some(cpu_ptr) == self.cpu_shared.as_ref().map(|obj| obj.ptr.as_ptr()) {
            self.allocate_cpu();
            copy_nonoverlapping(
                cpu_ptr,
                self.cpu_owned.as_ref().unwrap().ptr.as_ptr(),
                self.size,
            );
        }
        self.cpu_shared = None;
        #[cfg(feature = "cuda")]
        self.deallocate_cuda();
        cpu_ptr
    }

    #[cfg(feature = "cuda")]
    pub unsafe fn __cuda_ptr(&mut self) -> *mut u8 {
        if let Some(ptr) = self.get_cuda_ptr() {
            ptr.as_ptr()
        } else {
            let cpu_ptr = self.get_cpu_ptr().unwrap();
            let cuda_ptr = self.allocate_cuda();
            copy_cpu_to_cuda(cuda_ptr.as_ptr(), cpu_ptr.as_ptr(), self.size);
            cuda_ptr.as_ptr()
        }
    }

    #[cfg(feature = "cuda")]
    pub unsafe fn __cuda_ptr_mut(&mut self) -> *mut u8 {
        let cuda_ptr = self.__cuda_ptr();
        self.cpu_shared = None;
        self.cpu_exclusive = None;
        self.deallocate_cpu();
        cuda_ptr
    }

    pub unsafe fn __clone(&self) -> Self {
        Self {
            cpu_shared: self.cpu_shared.clone(),
            cpu_exclusive: self.cpu_exclusive.clone(),
            cpu_owned: self.cpu_owned.clone(),
            #[cfg(feature = "cuda")]
            cuda_owned: self.cuda_owned.clone(),
            size: self.size,
            id: self.id,
        }
    }

    pub unsafe fn __forget(&mut self) {
        self.cpu_owned = None;
        #[cfg(feature = "cuda")]
        {
            self.cuda_owned = None;
        }
    }

    pub unsafe fn __offset(&mut self, offset: u64, size: u64) {
        if let Some(obj) = self.cpu_shared.as_mut() {
            obj.offset += offset as usize;
        }
        if let Some(obj) = self.cpu_exclusive.as_mut() {
            obj.offset += offset as usize;
        }
        if let Some(obj) = self.cpu_owned.as_mut() {
            obj.offset += offset as usize;
        }
        #[cfg(feature = "cuda")]
        if let Some(obj) = self.cuda_owned.as_mut() {
            obj.offset += offset as usize;
        }
        self.size = size as usize;
    }

    pub unsafe fn __cmp_ids(&self, other: &HerculesBox<'_>) -> bool {
        self.id == other.id
    }
}

impl<'a> Drop for HerculesBox<'a> {
    fn drop(&mut self) {
        unsafe {
            self.deallocate_cpu();
            #[cfg(feature = "cuda")]
            self.deallocate_cuda();
        }
    }
}
