use std::collections::{HashMap, HashSet};
use std::fs::File;
use std::io::Read;

use lrlex::DefaultLexerTypes;
use lrpar::NonStreamingLexer;

use hercules_ir::ir::*;

mod parser;
use crate::parser::lexer;

// FunctionMap tracks a map from function numbers (produced by semantic analysis) to a tuple of
// - The map from label names to their numbers
// - The name of the function
// - A list of the instances of the function tracking
//   + The instantiated type variables
//   + The resulting FunctionID
//   + A list of each label, tracking the structure at the label and a set of
//     the labels which are its descendants
//   + A map from NodeID to the innermost label containing it
// This is the core data structure provided from code generation, along with the
// module
pub type FunctionMap = HashMap<
    usize,
    (
        HashMap<String, usize>,
        String,
        Vec<(
            Vec<TypeID>,
            FunctionID,
            Vec<(LabeledStructure, HashSet<usize>)>,
            HashMap<NodeID, usize>,
        )>,
    ),
>;
// LabeledStructure represents structures from the source code and where they
// exist in the IR
#[derive(Copy, Clone)]
pub enum LabeledStructure {
    Nothing(),
    Expression(NodeID),
    Loop(NodeID),   // Header
    Branch(NodeID), // If node
}

/*
pub fn schedule(module: &Module, info: FunctionMap, schedule: String) -> Result<Vec<Plan>, String> {
    if let Ok(mut file) = File::open(schedule) {
        let mut contents = String::new();
        if let Ok(_) = file.read_to_string(&mut contents) {
            let lexerdef = lexer::lexerdef();
            let lexer = lexerdef.lexer(&contents);
            let (res, errs) = parser::parse(&lexer);

            if errs.is_empty() {
                match res {
                    None => Err(format!("No parse errors, but no parsing failed")),
                    Some(schd) => {
                        let mut sched = generate_schedule(module, info, schd, &lexer)?;
                        let mut schedules = vec![];
                        for i in 0..sched.len() {
                            schedules.push(sched.remove(&FunctionID::new(i)).unwrap());
                        }
                        Ok(schedules)
                    }
                }
            } else {
                Err(errs
                    .iter()
                    .map(|e| format!("Syntax Error: {}", e.pp(&lexer, &parser::token_epp)))
                    .collect::<Vec<_>>()
                    .join("\n"))
            }
        } else {
            Err(format!("Unable to read input file"))
        }
    } else {
        Err(format!("Unable to open input file"))
    }
}

// a plan that tracks additional information useful while we construct the
// schedule
struct TempPlan {
    schedules: Vec<Vec<Schedule>>,
    // we track both the partition each node is in and what labeled caused us
    // to assign that partition
    partitions: Vec<(usize, PartitionNumber)>,
    partition_devices: Vec<Vec<Device>>,
}
type PartitionNumber = usize;

impl Into<Plan> for TempPlan {
    fn into(self) -> Plan {
        let num_partitions = self.partition_devices.len();
        Plan {
            schedules: self.schedules,
            partitions: self
                .partitions
                .into_iter()
                .map(|(_, n)| PartitionID::new(n))
                .collect::<Vec<_>>(),
            partition_devices: self
                .partition_devices
                .into_iter()
                .map(|mut d| {
                    if d.len() != 1 {
                        panic!("Partition with multiple devices")
                    } else {
                        d.pop().unwrap()
                    }
                })
                .collect::<Vec<_>>(),
            num_partitions: num_partitions,
        }
    }
}

fn generate_schedule(
    module: &Module,
    info: FunctionMap,
    schedule: Vec<parser::FuncDirectives>,
    lexer: &dyn NonStreamingLexer<DefaultLexerTypes<u32>>,
) -> Result<HashMap<FunctionID, Plan>, String> {
    let mut res: HashMap<FunctionID, TempPlan> = HashMap::new();

    // We initialize every node in every function as not having any schedule
    // and being in the default partition which is a CPU-only partition
    // (a result of label 0)
    for (_, (_, _, func_insts)) in info.iter() {
        for (_, func_id, _, _) in func_insts.iter() {
            let num_nodes = module.functions[func_id.idx()].nodes.len();
            res.insert(
                *func_id,
                TempPlan {
                    schedules: vec![vec![]; num_nodes],
                    partitions: vec![(0, 0); num_nodes],
                    partition_devices: vec![vec![Device::CPU]],
                },
            );
        }
    }

    // Construct a map from function names to function numbers
    let mut function_names: HashMap<String, usize> = HashMap::new();
    for (num, (_, nm, _)) in info.iter() {
        function_names.insert(nm.clone(), *num);
    }
    // Make the map immutable
    let function_names = function_names;

    for parser::FuncDirectives {
        span: _,
        func,
        directives,
    } in schedule
    {
        // Identify the function
        let parser::Func {
            span: _,
            name: func_name,
        } = func;
        let name = lexer.span_str(func_name).to_string();
        let func_num = match function_names.get(&name) {
            Some(num) => num,
            None => {
                return Err(format!("Function {} is undefined", name));
            }
        };

        // Identify label information
        let (label_map, _, func_inst) = info.get(func_num).unwrap();
        let get_label_num = |label_span| {
            let label_name = lexer.span_str(label_span).to_string();
            match label_map.get(&label_name) {
                Some(num) => Ok(*num),
                None => Err(format!("Label {} undefined in {}", label_name, name)),
            }
        };

        // Process the partitioning and scheduling directives for each instance
        // of the function
        for (_, func_id, label_info, node_labels) in func_inst {
            let func_info = res.get_mut(func_id).unwrap();

            for directive in &directives {
                match directive {
                    parser::Directive::Partition {
                        span: _,
                        labels,
                        devices,
                    } => {
                        // Setup the new partition
                        let partition_num = func_info.partition_devices.len();
                        let mut partition_devices = vec![];

                        for parser::Device { span: _, name } in devices {
                            let device_name = lexer.span_str(*name).to_string();
                            if device_name == "cpu" {
                                partition_devices.push(Device::CPU);
                            } else if device_name == "gpu" {
                                partition_devices.push(Device::GPU);
                            } else {
                                return Err(format!("Invalid device {}", device_name));
                            }
                        }

                        func_info.partition_devices.push(partition_devices);

                        for label in labels {
                            let label_num = get_label_num(*label)?;
                            let descendants = &label_info[label_num].1;

                            node_labels
                                .iter()
                                .filter_map(|(node, label)| {
                                    if *label == label_num || descendants.contains(label) {
                                        Some(node.idx())
                                    } else {
                                        None
                                    }
                                })
                                .for_each(|node| {
                                    let node_part: &mut (usize, PartitionNumber) =
                                        &mut func_info.partitions[node];
                                    if !descendants.contains(&node_part.0) {
                                        *node_part = (label_num, partition_num);
                                    }
                                });
                        }
                    }
                    parser::Directive::Schedule {
                        span: _,
                        command,
                        args,
                    } => {
                        let command = lexer.span_str(*command).to_string();
                        if command == "parallelize" {
                            for label in args {
                                let label_num = get_label_num(*label)?;
                                match label_info[label_num].0 {
                                    LabeledStructure::Loop(header) => {
                                        func_info.schedules[header.idx()]
                                            .push(Schedule::ParallelReduce);
                                    }
                                    _ => {
                                        return Err(format!(
                                            "Cannot parallelize {}, not a loop",
                                            lexer.span_str(*label)
                                        ));
                                    }
                                }
                            }
                        } else if command == "vectorize" {
                            for label in args {
                                let label_num = get_label_num(*label)?;
                                match label_info[label_num].0 {
                                    LabeledStructure::Loop(header) => {
                                        // FIXME: Take the factor as part of schedule
                                        func_info.schedules[header.idx()]
                                            .push(Schedule::Vectorizable(8));
                                    }
                                    _ => {
                                        return Err(format!(
                                            "Cannot vectorize {}, not a loop",
                                            lexer.span_str(*label)
                                        ));
                                    }
                                }
                            }
                        } else {
                            return Err(format!("Command {} undefined", command));
                        }
                    }
                }
            }
            /*

            /*
            for parser::Command { span : _, name : command_name,
                                  args : command_args } in commands.iter() {
                if command_args.len() != 0 { todo!("Command arguments not supported") }

                let command = lexer.span_str(*command_name).to_string();
                if command == "cpu" || command == "gpu" {
                    let partition = res.get(func_id).unwrap()
                                       .partition_devices.len();
                    res.get_mut(func_id).unwrap().partition_devices.push(
                        if command == "cpu" { Device::CPU }
                        else { Device::GPU });

                    node_labels.iter()
                               .filter_map(|(node, label)|
                                   if label_num == *label
                                   || label_info[label_num].1.contains(&label) {
                                       Some(node.idx())
                                   } else {
                                       None
                                   })
                               .for_each(|node| {
                                     let node_part : &mut (usize, PartitionNumber) =
                                        &mut res.get_mut(func_id).unwrap().partitions[node];
                                     if !label_info[label_num].1.contains(&node_part.0) {
                                        *node_part = (label_num, partition);
                                     }});
                } else if command == "parallel" || command == "vectorize" {
                    match label_info[label_num].0 {
                        LabeledStructure::Loop(header) => {
                            res.get_mut(func_id).unwrap()
                                .schedules[header.idx()]
                                .push(if command == "parallel" {
                                    Schedule::ParallelReduce
                                } else {
                                    Schedule::Vectorize
                                });
                        },
                        _ => {
                            return Err(format!("Cannot parallelize, not a loop"));
                        },
                    }
                } else {
                    return Err(format!("Command {} undefined", command));
                }
            }
            */

            func_info.partition_devices.push(partition_devices);
            */
        }
    }

    Ok(res
        .into_iter()
        .map(|(f, p)| (f, p.into()))
        .collect::<HashMap<_, _>>())
}
*/
